#include <algorithm>
#include <mutex>

#include "ArmarXCore/core/logging/Logging.h"
#include "ArmarXCore/core/services/tasks/PeriodicTask.h"

#include "RobotAPI/libraries/armem/core/MemoryID.h"
#include "RobotAPI/libraries/armem/core/Time.h"
#include "RobotAPI/libraries/armem/core/wm/memory_definitions.h"
#include <RobotAPI/libraries/armem/client/MemoryNameSystem.h>
#include <RobotAPI/libraries/armem/client/query/Builder.h>

#include <armarx/navigation/client/services/MemorySubscriber.h>
#include <armarx/navigation/core/events.h>


namespace armarx::navigation::client
{

    MemorySubscriber::MemorySubscriber(const std::string& callerId,
                                       armem::client::MemoryNameSystem& mns) :
        callerId(callerId), memoryNameSystem(mns), lastMemoryPoll(armem::Time::now())
    {
        ARMARX_TRACE;
        memoryReader = memoryNameSystem.getReader(armem::MemoryID().withMemoryName("Navigation"));

        ARMARX_INFO << "MemorySubscriber: will handle all events newer than " << lastMemoryPoll
                    << ".";

        // subscription api
        armem::MemoryID subscriptionID;
        subscriptionID.coreSegmentName = "Events";
        subscriptionID.providerSegmentName = callerId;

        // TODO add: memoryNameSystem.subscribe(subscriptionID, this, &MemorySubscriber::onEntityUpdate);

        // polling api
        ARMARX_TRACE;
        task = new armarx::PeriodicTask<MemorySubscriber>(
            this, &MemorySubscriber::runPollMemoryEvents, 20);
        task->start();
    }

    void
    MemorySubscriber::onEntityUpdate(const armem::MemoryID& subscriptionID,
                                     const std::vector<armem::MemoryID>& snapshotIDs)
    {
        ARMARX_INFO << "Received " << snapshotIDs.size() << " events from memory";
        const armem::client::QueryResult qResult = memoryReader.queryMemoryIDs(snapshotIDs);

        if (not qResult.success) /* c++20 [[unlikely]] */
        {
            ARMARX_WARNING << deactivateSpam(0.1F) << "Memory lookup failed.";
            return;
        }

        handleEvents(qResult.memory);
    }


    void
    MemorySubscriber::handleEvent(const armem::wm::EntityInstance& memoryEntity)
    {
        ARMARX_TRACE;

        // only one event will be handled at a time
        std::lock_guard g{eventHandlingMtx};

        ARMARX_CHECK(memoryEntity.id().hasEntityName());
        const auto& eventName = memoryEntity.id().entityName;

        ARMARX_IMPORTANT << "Handling event `" << eventName << "`";

        // factory
        if (eventName == "GlobalPlanningFailedEvent")
        {
            core::GlobalPlanningFailedEvent evt;
            // fromAron(memoryEntity.data(), evt);

            globalPlanningFailed(evt);
        }
        else if (eventName == "MovementStartedEvent")
        {
            core::MovementStartedEvent evt;
            // fromAron(memoryEntity.data(), evt);

            movementStarted(evt);
        }
        else if (eventName == "GoalReachedEvent")
        {
            core::GoalReachedEvent evt;
            // fromAron(memoryEntity.data(), evt);

            goalReached(evt);
        }
        else if (eventName == "WaypointReachedEvent")
        {
            core::WaypointReachedEvent evt;
            // fromAron(memoryEntity.data(), evt);

            waypointReached(evt);
        }
        else if (eventName == "SafetyThrottlingTriggeredEvent")
        {
            core::SafetyThrottlingTriggeredEvent evt;
            // fromAron(memoryEntity.data(), evt);

            safetyThrottlingTriggered(evt);
        }
        else if (eventName == "SafetyStopTriggeredEvent")
        {
            core::SafetyStopTriggeredEvent evt;
            // fromAron(memoryEntity.data(), evt);

            safetyStopTriggered(evt);
        }
        else if (eventName == "UserAbortTriggeredEvent")
        {
            core::UserAbortTriggeredEvent evt;
            // fromAron(memoryEntity.data(), evt);

            userAbortTriggered(evt);
        }
        else if (eventName == "InternalErrorEvent")
        {
            core::InternalErrorEvent evt;
            // fromAron(memoryEntity.data(), evt);

            internalError(evt);
        }
        else
        {
            ARMARX_WARNING << "Unknown event `" << eventName << "`";
        }
    }

    MemorySubscriber::~MemorySubscriber()
    {
        ARMARX_INFO << "Stopping event polling";
        task->stop();
        ARMARX_INFO << "done.";
    }


    void
    MemorySubscriber::runPollMemoryEvents()
    {
        ARMARX_TRACE;

        const armem::Time now = armem::Time::now();

        armem::client::query::Builder qb;
        // clang-format off
        qb
        .coreSegments().withName("Events")
        .providerSegments().withName(callerId)
        .entities().all()
        .snapshots().timeRange(lastMemoryPoll, now);
        // clang-format on

        // ARMARX_DEBUG << "Polling memory events in interval "
        //              << "[" << lastMemoryPoll << ", " << now << "]";

        lastMemoryPoll = now;

        const armem::client::QueryResult qResult = memoryReader.query(qb.buildQueryInput());

        // ARMARX_DEBUG << "Lookup result in reader: " << qResult;

        if (not qResult.success) /* c++20 [[unlikely]] */
        {
            ARMARX_WARNING << deactivateSpam(0.1F) << "Memory lookup failed.";
            return;
        }

        handleEvents(qResult.memory);
    }

    void
    MemorySubscriber::handleEvents(const armem::wm::Memory& memory)
    {
        ARMARX_TRACE;

        // clang-format off
        const armem::wm::CoreSegment& coreSegment = memory
                .getCoreSegment("Events");
        // clang-format on

        std::vector<armem::wm::EntityInstance> events;

        coreSegment.forEachInstance([&](const armem::wm::EntityInstance& instance)
                                    { events.push_back(instance); });

        const auto sortByTimestampAsc = [](const armem::wm::EntityInstance& a,
                                           const armem::wm::EntityInstance& b) -> bool
        {
            ARMARX_CHECK(a.id().hasTimestamp());
            ARMARX_CHECK(b.id().hasTimestamp());

            return a.id().timestamp < b.id().timestamp;
        };

        std::sort(events.begin(), events.end(), sortByTimestampAsc);

        std::for_each(events.begin(), events.end(), [&](const auto& event) { handleEvent(event); });
    }


} // namespace armarx::navigation::client
