#include "Visualization.h"

#include <chrono>
#include <thread>

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/time/TimeUtil.h>


namespace armarx::navigation::util
{

    Visualization::Visualization(armarx::viz::Client& arviz, const Params& params) :
        params(params), arviz(arviz)
    {
        robotVis = viz::Robot("robot").file("", params.robotModelFileName);
        robotVis.useCollisionModel();
        robotVis.hide();

        arviz.commitLayerContaining("target", robotVis);
    }

    void
    Visualization::visualizePath(std::vector<Eigen::Vector3f>& path)
    {
        // TODO(fabian.reister): more finegrained locking
        std::lock_guard g{paramsMutex};

        if (!params.enableVisualization)
        {
            return;
        }

        if (path.size() < 2)
        {
            ARMARX_WARNING << "Unable to visualize path with less than two segments";
            return;
        }

        viz::Layer pathLayer = arviz.layer("path");
        for (size_t i = 1; i < path.size(); i++)
        {
            Eigen::Vector3f prevPose = Eigen::Vector3f::Zero();
            prevPose.head<2>() = path.at(i - 1).head<2>();
            Eigen::Vector3f currentPose = Eigen::Vector3f::Zero();
            currentPose.head<2>() = path.at(i).head<2>();
            viz::Arrow a = viz::Arrow("path_segment_" + std::to_string(i));
            a.color(viz::Color::orange());
            a.width(25);
            a.fromTo(prevPose, currentPose);
            pathLayer.add(a);
        }

        arviz.commit({pathLayer});

        visualizeTrajectory(path);
    }


    void
    Visualization::visualizeTrajectory(std::vector<Eigen::Vector3f>& path)
    {
        if (!params.enableTrajectoryVisualization)
        {
            return;
        }

        viz::Layer robotLayer = arviz.layer("trajectory");

        robotVis.overrideColor(viz::Color::white(params.alphaValue));
        robotVis.show();
        robotLayer.add(robotVis);

        for (size_t i = 1; i < path.size(); i++)
        {
            const Eigen::Vector3f& prevPose = path.at(i - 1);
            const Eigen::Vector3f& currentPose = path.at(i);

            Eigen::Vector3f diff = (currentPose - prevPose);

            const float steps = std::ceil(diff.head<2>().norm() / 10.0f);
            diff /= steps;

            std::lock_guard<std::mutex> lock(mutex);

            for (size_t j = 0; j < steps; j++)
            {
                Eigen::Vector2f c = prevPose.head<2>() + diff.head<2>() * j;
                float alpha = prevPose.z() + diff.z() * j;

                Eigen::Affine3f m(Eigen::Affine3f::Identity());
                m.translate(Eigen::Vector3f(c.x(), c.y(), 0.0));
                m.rotate(Eigen::AngleAxisf(alpha, Eigen::Vector3f::UnitZ()));

                robotVis.pose(m.matrix());

                arviz.commit({robotLayer});

                std::this_thread::sleep_for(std::chrono::milliseconds(10));
            }
        }

        robotVis.hide();

        std::lock_guard<std::mutex> lock(mutex);

        arviz.commit({robotLayer});
    }


    void
    Visualization::asyncHideDelayed()
    {
        std::thread{
            [this]
            {
                std::this_thread::sleep_for(std::chrono::seconds(params.hideDelaySeconds));

                std::lock_guard<std::mutex> lock(mutex);

                if (lastUpdate <= (TimeUtil::GetTime().toSeconds() - params.hideDelaySeconds))
                {
                    robotVis.hide();
                    arviz.commitLayerContaining("target", robotVis);
                    arviz.commit(arviz.layer("path"));
                }
            }}
            .detach();
    }

    void
    Visualization::success()
    {
        // TODO(fabian.reister): more finegrained locking
        std::lock_guard g{paramsMutex};

        if (!params.enableVisualization)
        {
            return;
        }

        robotVis.overrideColor(viz::Color::green(params.alphaValue));
        {
            std::lock_guard<std::mutex> lock(mutex);
            arviz.commitLayerContaining("target", robotVis);
        }
        asyncHideDelayed();
    }

    void
    Visualization::failed()
    {
        // TODO(fabian.reister): more finegrained locking
        std::lock_guard g{paramsMutex};

        if (!params.enableVisualization)
        {
            return;
        }
        robotVis.overrideColor(viz::Color::red(params.alphaValue));
        {
            std::lock_guard<std::mutex> lock(mutex);
            arviz.commitLayerContaining("target", robotVis);
        }

        asyncHideDelayed();
    }

    void
    Visualization::setTarget(const core::Pose& m)
    {
        // TODO(fabian.reister): more finegrained locking
        std::lock_guard g{paramsMutex};

        if (!params.enableVisualization)
        {
            return;
        }

        robotVis.show();
        robotVis.overrideColor(viz::Color::azure(128, params.alphaValue));
        robotVis.pose(m.matrix());

        std::lock_guard<std::mutex> lock(mutex);
        arviz.commitLayerContaining("target", robotVis);

        lastUpdate = TimeUtil::GetTime().toSeconds();
    }


} // namespace armarx::navigation::util
